import datetime
import numpy as np
from paux import settings
from paux.candle.history import load as _load
from paux.candle.history import save as _save
from paux import param as _param
from typing import Literal, Union

settings_data = settings.read_settings()

__CANDLE_BASE_DIR = settings_data['CANDLE_OKX_BASE_DIR']
__CANDLE_TIMEZONE = settings_data['CANDLE_OKX_TIMEZONE']


# 读取candle
def load_candle_by_date(
        instType: str,
        instId: str,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir:str =__CANDLE_BASE_DIR,
        timezone: str = __CANDLE_TIMEZONE,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        valid_interval=True,
        valid_start=True,
        valid_end=True,
        minus_bar:int = 1,
):
    return _load.load_candle_by_date(**_param.to_local(locals()))


# 按照日期读取candle_map
def load_candle_map_by_date(
        instType: str,
        instIds: list,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: str = __CANDLE_BASE_DIR,
        timezone: str = __CANDLE_TIMEZONE,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        contains='-USDT',
        p_num=4,
        valid_interval=True,
        valid_start=True,
        valid_end=True,
        minus_bar:int = 1,
):
    return _load.load_candle_map_by_date(**_param.to_local(locals()))


# 通过文件地址读取Candle
def load_candle_by_file(
        instType: str,
        instId: str,
        path: str = None,
        base_dir: str = __CANDLE_BASE_DIR,
        timezone: str = __CANDLE_TIMEZONE,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        valid_interval: bool = True,
):
    return _load.load_candle_by_file(**_param.to_local(locals()))


# 通过文件夹地址读取Candle_map
def load_candle_map_by_file(
        instType: str,
        instIds: list = [],
        path: str = None,
        base_dir: str = __CANDLE_BASE_DIR,
        timezone: str = __CANDLE_TIMEZONE,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        valid_interval: bool = True,

):
    return _load.load_candle_map_by_file(**_param.to_local(locals()))


# 按照日期保存Candle
def save_candle_by_date(
        candle: np.array,
        instType: str,
        instId: str,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: str = __CANDLE_BASE_DIR,
        timezone: str = __CANDLE_TIMEZONE,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        drop_duplicate: bool = True,
        sort: bool = True,
        valid_interval: bool = True,
        valid_start: bool = True,
        valid_end: bool = True,
        minus_bar:int = 1,
):
    return _save.save_candle_by_date(**_param.to_local(locals()))


# 按照日期保存candle_map
def save_candle_map_by_date(
        candle_map: dict,
        instType: str,
        instIds: list,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: str = __CANDLE_BASE_DIR,
        timezone: str = __CANDLE_TIMEZONE,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        drop_duplicate: bool = True,
        sort: bool = True,
        valid_interval: bool = True,
        valid_start: bool = True,
        valid_end: bool = True,
        minus_bar:int = 1,
):
    return _save.save_candle_map_by_date(**_param.to_local(locals()))


# 按照文件地址保存Candle
def save_candle_by_file(
        candle: np.array,
        instType: str,
        instId: str,
        path: str = None,
        base_dir: str = __CANDLE_BASE_DIR,
        timezone: str = __CANDLE_TIMEZONE,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        sort=True,
        drop_duplicate=True,
        valid_interval=True,
):
    return _save.save_candle_by_file(**_param.to_local(locals()))


# 按照文件地址保存Candle_map
def save_candle_map_by_file(
        candle_map: dict,
        instType: str,
        instIds: list,
        base_dir: str = __CANDLE_BASE_DIR,
        timezone: str = __CANDLE_TIMEZONE,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        sort: bool = True,
        drop_duplicate: bool = True,
        valid_interval: bool = True,
):
    return _save.save_candle_map_by_file(**_param.to_local(locals()))



import numpy as np
from paux import exception

# 得到时间间隔，单位毫秒
def get_interval(bar: str, MINUTE_BAR_INTERVAL=60000) -> float:
    '''

    Get ts interval by bar

    :param bar:
    Time interval of historical k line
        bar = '1m'  1 minute
        bar = '5m'  5 minute
        bar = '1h'  1 hour
        bar = '1d'  1 day
        ... ...

    :param MINUTE_BAR_INTERVAL:
    Number interval per minute, in milliseconds
        MINUTE_BAR_INTERVAL = 60000    Ts is in milliseconds
        MINUTE_BAR_INTERVAL = 60       Ts is in seconds

    :return: interval
    '''
    bar_int = int(bar[0:-1].strip())
    suffix = bar[-1].lower()
    if suffix == 'm':
        interval = MINUTE_BAR_INTERVAL * bar_int
    elif suffix == 'h':
        interval = MINUTE_BAR_INTERVAL * 60 * bar_int
    elif suffix == 'd':
        interval = MINUTE_BAR_INTERVAL * 60 * 24 * bar_int
    elif suffix == 'w':
        interval = MINUTE_BAR_INTERVAL * 60 * 24 * 7 * bar_int
    else:
        raise exception.ParamException(
            "The format of bar should be like ['1m ',' 2m ',' 1h ',' 2h ',' 1d ',' 2d '...]"
        )
    return interval


# 估计K线中的时间粒度间隔
def predict_interval(candle: np.array) -> float:
    '''

    Predict ts interval by candle

    :param candle: History Candle
        array[
            [ts,open,high,low,close,volume],
            [ts,open,high,low,close,volume],
            [ts,open,high,low,close,volume],
        ]

    :return:interval
    '''
    return np.min(np.diff(candle[:, 0]))


class ParamException(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class PredictException(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class CandleFileError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class CandleDiffError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class CandleStartError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class CandleEndError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class PlatformError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class UnexpectError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


from paux.candle.history import candle_okx as OKX
from paux.candle.history import candle_ib as IB
from paux.candle.history import candle_tda as TDA
from paux.candle.history import candle_huobi as HUOBI
from paux.candle.history import candle_binance as BINANCE
from paux.candle.history import candle_binance as BINANCE


from pprint import pprint as _pprint


# 逐行打印内容
def pprint(*args):
    for arg in args:
        if type(arg) in [int, float, str]:
            print(arg)
        else:
            _pprint(arg)
    print('')



'''
与平台相关的函数
'''

import platform
import os


def get_platform():
    sys_platform = platform.platform().upper()
    sys_names = (
        'WINDOWS', 'MACOS', 'LINUX'

    )
    for sys_name in sys_names:
        if sys_name in sys_platform:
            return sys_name
    return 'OTHER'


def get_maxOs_documents_path():
    root_path = '/Users'
    filepaths = []
    for filename in os.listdir(root_path):
        if filename.startswith('.'):
            continue
        if filename == 'Shared':
            continue
        filepath = os.path.join(root_path, filename)
        if os.path.isdir(filepath) and 'Documents' in os.listdir(filepath):
            filepaths.append(filepath)

    if not filepaths:
        raise Exception('没有找到MacOs用户名')
    documents_path = os.path.join(
        filepaths[0], 'Documents'
    )
    return documents_path


if __name__ == "__main__":
    print(get_platform())
    print(get_maxOs_documents_path())



import os
import pandas as pd
import datetime
from typing import Union, Literal
from paux.candle.history import lite as _lite
from paux.candle import transform as _transform
from paux.candle import valid as _valid
from paux.candle import interval as _interval
from paux import param as _param
from paux import process as _process
from paux import date as _date
from paux import exception


# 读取candle
def load_candle_by_date(
        instType: str,
        instId: str,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: str,
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        valid_interval=True,
        valid_start=True,
        valid_end=True,
        minus_bar=1,
):
    if minus_bar:
        end = _date.to_ts(end, timezone) - _interval.get_interval(bar=bar) * minus_bar
    # 检查是否有数据
    if not _lite.check_candle_day_path(
            instType=instType, instId=instId, start=start,
            end=end, timezone=timezone, bar=bar, base_dir=base_dir
    ):
        error_msg = 'candle数据不足'
        raise exception.CandleFileError(error_msg)

    # 获取数据的序列
    date_range = _date.get_range_dates(
        start=start,
        end=end,
        timezone=timezone,
    )
    # 获取数据的路径列表
    paths = [
        _lite.get_candle_day_path(
            instType=instType,
            instId=instId,
            date=date,
            timezone=timezone,
            bar=bar,
            base_dir=base_dir,
        )
        for date in date_range
    ]
    # 读取数据
    dfs = []
    for path in paths:
        df = pd.read_csv(path)
        dfs.append(df)
    # 合并数据
    candle = _transform.concat_candle(candles=dfs, drop_duplicate=True, sort=True)
    if valid_interval and not _valid.valid_interval(candle, bar=bar):
        msg = '{instId}的candle的时间间隔错误'.format(instId=instId)
        raise exception.CandleDiffError(msg)
    if valid_start and not _valid.valid_start(candle=candle, start=start, timezone=timezone):
        msg = '{instId}的candle的时间起始错误'.format(instId=instId)
        raise exception.CandleStartError(msg)
    end = _date.to_datetime(end, timezone)
    if valid_end and not _valid.valid_end(candle=candle, end=end, timezone=timezone):
        msg = '{instId}的candle的时间终止错误'.format(instId=instId)
        raise exception.CandleEndError(msg)
    return candle


'''
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: str,
        timezone: str = None,
'''


# 按照日期读取candle_map
def load_candle_map_by_date(
        instType: str,
        instIds: list,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: str,
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        contains='-USDT',
        p_num=4,
        valid_interval=True,
        valid_start=True,
        valid_end=True,
        minus_bar=1,
):
    if minus_bar:
        end = _date.to_ts(end, timezone) - _interval.get_interval(bar=bar) * minus_bar
    # 如果没有产品的名字，获取产品类型数据中，有start_date到end_date中有完整数据的instId
    if not instIds:
        dates = _date.get_range_dates(start=start, end=end, timezone=timezone)
        instIds = set()
        for date in dates:
            date_dirpath = os.path.dirname(
                _lite.get_candle_day_path(
                    base_dir=base_dir, instType=instType, timezone=timezone,
                    date=date, bar=bar, instId=''
                )
            )
            this_instIds = []
            for filename in os.listdir(date_dirpath):
                instId = filename.rsplit('.', maxsplit=1)[0]
                if contains and not contains in instId:
                    continue
                else:
                    this_instIds.append(instId)
            if not instIds:
                instIds = set(this_instIds)
            else:
                instIds = instIds & set(this_instIds)
    instIds = list(instIds)
    candle_map = {}
    if p_num > 1:
        params = []
        for instId in instIds:
            params.append(
                dict(
                    instType=instType,
                    instId=instId,
                    start=start,
                    end=end,
                    timezone=timezone,
                    bar=bar,
                    base_dir=base_dir,
                    valid_interval=valid_interval,
                    valid_start=valid_start,
                    valid_end=valid_end,
                    minus_bar=0,
                )
            )
        results = _process.pool_worker(
            params=params,
            p_num=p_num,
            func=load_candle_by_date,
            skip_exception=False,
        )
        for i, result in enumerate(results):
            if _param.isnull(result):
                continue
            instId = instIds[i]
            candle_map[instId] = result

    else:
        for instId in instIds:
            candle_map[instId] = load_candle_by_date(
                instType=instType,
                instId=instId,
                start=start,
                end=end,
                timezone=timezone,
                bar=bar,
                base_dir=base_dir,
                valid_interval=valid_interval,
                valid_start=valid_start,
                valid_end=valid_end,
            )
    # candle_map排序
    candle_map_sorted = {}
    instIds = sorted([instId for instId in candle_map.keys()])
    for instId in instIds:
        candle_map_sorted[instId] = candle_map[instId]

    for instId, candle in candle_map_sorted.items():
        if valid_interval and not _valid.valid_interval(candle, bar=bar):
            msg = '{instId}的candle的时间间隔错误'.format(instId=instId)
            raise exception.CandleDiffError(msg)
        if valid_start and not _valid.valid_start(candle=candle, start=start, timezone=timezone):
            msg = '{instId}的candle的时间起始错误'.format(instId=instId)
            raise exception.CandleStartError(msg)
        if valid_end and not _valid.valid_end(candle=candle, end=end, timezone=timezone):
            msg = '{instId}的candle的时间终止错误'.format(instId=instId)
            raise exception.CandleEndError(msg)
    return candle_map_sorted


# 通过文件地址读取Candle
def load_candle_by_file(
        instType: str,
        instId: str,
        path: str = None,
        base_dir: str = '',
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        valid_interval: bool = True,
):
    '''
    如果有path路径，按照path路径读取文件
    如果没有path路径，按照base_dir、instId、instType、bar和timezone计算产品路径
    '''
    if path == None:
        path = _lite.get_candle_file_path(
            instType=instType,
            instId=instId,
            bar=bar,
            timezone=timezone,
            base_dir=base_dir,
        )
    df = pd.read_csv(path)
    candle = _transform.to_candle(candle=df, drop_duplicate=True, sort=True)
    if valid_interval and not _valid.valid_interval(candle, bar=bar):
        msg = '{instId}的candle的时间间隔错误'.format(instId=instId)
        return exception.CandleDiffError(msg)
    return candle


# 通过文件夹地址读取Candle_map
def load_candle_map_by_file(
        instType: str,
        instIds: list = [],
        path: str = None,
        base_dir: str = '',
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        valid_interval: bool = True,

):
    # 如果没有文件夹的地址
    if path == None:
        path = os.path.dirname(
            _lite.get_candle_file_path(
                instType=instType,
                base_dir=base_dir,
                timezone=timezone,
                bar=bar,
                instId='xx',
            )
        )
    # 如果没有产品的名字，读取文件夹中全部的产品
    if not instIds:
        filenames = os.listdir(path)
        instIds = []
        for filename in filenames:
            instId = filename.rsplit('.', maxsplit=1)[0]
            instIds.append(instId)
    # 读取数据
    candle_map = {}
    for instId in instIds:
        candle = load_candle_by_file(
            instId=instId,
            instType=instType,
            base_dir=base_dir,
            timezone=timezone,
            bar=bar,
            valid_interval=valid_interval
        )
        candle_map[instId] = candle
    # candle_map排序
    candle_map_sorted = {}
    instIds = sorted([instId for instId in candle_map.keys()])
    for instId in instIds:
        candle_map_sorted[instId] = candle_map[instId]
    return candle_map_sorted



import os
import numpy as np
import pandas as pd
import datetime
from typing import Union, Literal
from paux.candle.history import lite as _lite
from paux.candle import transform as _transform
from paux.candle import valid as _valid
from paux.candle import interval as _interval
from paux import date as _date
from paux import exception


# 按照日期保存Candle
def save_candle_by_date(
        candle: np.array,
        instType: str,
        instId: str,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: str,
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        drop_duplicate: bool = True,
        sort: bool = True,
        valid_interval: bool = True,
        valid_start: bool = True,
        valid_end: bool = True,
        minus_bar=1,
):
    '''
    边按照日期写入，边进行valid，如果valid报告错误，之前的数据可以成功写入，后面的数据则不会继续写入
    '''
    if minus_bar:
        end = _date.to_ts(end, timezone) - _interval.get_interval(bar=bar) * minus_bar
    # 如果需要去重或者需要排序
    if drop_duplicate or sort:
        candle = _transform.to_candle(candle, drop_duplicate=True, sort=True)
    dates = _date.get_range_dates(start=start, end=end, timezone=timezone)
    # 验证数据
    for date in dates:
        start_ts = _date.to_ts(date=date, timezone=timezone)
        end_ts = start_ts + 1000 * 60 * 60 * 24 - _interval.get_interval(bar=bar)
        candle_date = candle[
            (candle[:, 0] >= start_ts) & (candle[:, 0] <= end_ts)
            ]
        if valid_interval and not _valid.valid_interval(candle_date, bar=bar):
            msg = '{instId}的candle的时间间隔错误'.format(instId=instId)
            raise exception.CandleDiffError(msg)
        if valid_start and not _valid.valid_start(candle=candle_date, start=start_ts, timezone=timezone):
            msg = '{instId}的candle的时间起始错误'.format(instId=instId)
            raise exception.CandleStartError(msg)
        if valid_end and not _valid.valid_end(candle=candle_date, end=end_ts, timezone=timezone):
            msg = '{instId}的candle的时间终止错误'.format(instId=instId)
            raise exception.CandleEndError(msg)
        path = _lite.get_candle_day_path(
            instType=instType,
            instId=instId,
            date=date,
            bar=bar,
            timezone=timezone,
            base_dir=base_dir
        )
        dirpath = os.path.dirname(path)
        if not os.path.isdir(dirpath):
            os.makedirs(dirpath)
        df_date = pd.DataFrame(candle_date)
        df_date.to_csv(path, index=False)
    return True


# 按照日期保存candle_map
def save_candle_map_by_date(
        candle_map: dict,
        instType: str,
        instIds: list,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: str = '',
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        drop_duplicate: bool = True,
        sort: bool = True,
        valid_interval: bool = True,
        valid_start: bool = True,
        valid_end: bool = True,
        minus_bar=1,
):
    '''
    如果写入的时候出现了错误，报错之前写入成功，报错后面的则不能正常写入
    '''
    if minus_bar:
        end = _date.to_ts(end, timezone) - _interval.get_interval(bar=bar) * minus_bar
    if not instIds:
        instIds = [instId for instId in candle_map.keys()]
    for instId in instIds:
        candle = candle_map[instId]
        save_candle_by_date(
            instType=instType,
            instId=instId,
            candle=candle,
            start=start,
            end=end,
            bar=bar,
            timezone=timezone,
            base_dir=base_dir,
            drop_duplicate=drop_duplicate,
            sort=sort,
            valid_interval=valid_interval,
            valid_start=valid_start,
            valid_end=valid_end,
            minus_bar=minus_bar
        )


# 按照文件地址保存Candle
def save_candle_by_file(
        candle: np.array,
        instType: str,
        instId: str,
        path: str = None,
        base_dir='',
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        sort=True,
        drop_duplicate=True,
        valid_interval=True,
):
    # 验证间隔
    if valid_interval and not _valid.valid_interval(candle=candle, bar=bar):
        msg = '{instId}的candle的时间间隔错误'.format(instId=instId)
        raise exception.CandleDiffError(msg)
    # 得到路径
    if path == None:
        path = _lite.get_candle_file_path(
            instType=instType,
            instId=instId,
            bar=bar,
            timezone=timezone,
            base_dir=base_dir,
        )
    # 排序与去重
    if sort or valid_interval:
        candle = _transform.to_candle(candle=candle, drop_duplicate=drop_duplicate, sort=sort)
    # 文件夹与路径
    dirpath = os.path.dirname(path)
    if not os.path.isdir(dirpath):
        os.makedirs(dirpath)
    df = pd.DataFrame(candle)
    # 写入文件
    df.to_csv(path, index=False)


# 按照文件地址保存Candle_map
def save_candle_map_by_file(
        candle_map: dict,
        instType: str,
        instIds: list,
        base_dir: str,
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        sort: bool = True,
        drop_duplicate: bool = True,
        valid_interval: bool = True,
):
    if not instIds:
        instIds = [instId for instId in candle_map.keys()]

    for instId in instIds:
        candle = candle_map[instId]
        save_candle_by_file(
            candle=candle,
            instType=instType,
            instId=instId,
            path=None,
            base_dir=base_dir,
            timezone=timezone,
            bar=bar,
            sort=sort,
            drop_duplicate=drop_duplicate,
            valid_interval=valid_interval,
        )



'''
get_setting_filepath    获取配置文件的路径
read_settings           读取配置文件
save_settings           保存配置文件
'''

import os
from paux import exception
from paux import system as _system
import json


# 初始化settings文件中的内容

def init_settings(settings_path=None):
    CANDLE_DIRNAME = 'CANDLE_DATA'
    platform = _system.get_platform()
    if platform == 'WINDOWS':
        CANDLE_BASE_DIR = os.path.join(os.getenv("SystemDrive"), CANDLE_DIRNAME)
    elif platform == 'MACOS':
        CANDLE_BASE_DIR = os.path.join(_system.get_maxOs_documents_path(), CANDLE_DIRNAME)
    elif platform == 'LINUX':
        CANDLE_BASE_DIR = os.path.join('/root/', CANDLE_DIRNAME)
    else:
        error_msg = f'未知系统 {platform}'
        raise exception.PlatformError(error_msg)

    init_data = {
        'CANDLE_OKX_BASE_DIR': os.path.join(CANDLE_BASE_DIR, 'OKX'),
        'CANDLE_OKX_TIMEZONE': 'Asia/Shanghai',
        'CANDLE_BINANCE_BASE_DIR': os.path.join(CANDLE_BASE_DIR, 'BINANCE'),
        'CANDLE_BINANCE_TIMEZONE': 'America/New_York',
        'CANDLE_HUOBI_BASE_DIR': os.path.join(CANDLE_BASE_DIR, 'HUOBI'),
        'CANDLE_HUOBI_TIMEZONE': 'Asia/Shanghai',
        'CANDLE_TDA_BASE_DIR': os.path.join(CANDLE_BASE_DIR, 'TDA'),
        'CANDLE_TDA_TIMEZONE': 'America/New_York',
        'CANDLE_IB_BASE_DIR': os.path.join(CANDLE_BASE_DIR, 'IB'),
        'CANDLE_IB_TIMEZONE': 'America/New_York',
        'CANDLE_EAST_MONEY_BAST_DIR': os.path.join(CANDLE_BASE_DIR, 'MONEY_EAST'),
        'CANDLE_EAST_MONEY_TIMEZONE': 'Asia/Shanghai',
    }
    # for key, value in init_data.items():
    #     msg = f'{key} = {value} (Y|N)\n'
    #     command = input(msg)
    #     if command.lower() == 'y':
    #         pass
    #     elif command.lower() == 'n':
    #         msg = f'{key} : \n'
    #         value = input(msg)
    #         init_data[key] = value
    save_settings(data=init_data, settings_path=settings_path)


# 获取配置文件路径
def get_settings_filepath(
        dirpath: str = None, filename: str = 'SETTINGS.JSON'
):
    '''
    :param dirpath: 配置文件的文件夹地址，如果没有本项目会使用paux的路径作为dirpath
    :param filename: 配置文件的名字
    :return: 配置文件路径
    '''
    if not dirpath:
        dirpath = os.path.dirname(os.path.dirname(__file__))
    elif not os.path.isdir(dirpath):
        raise exception.ParamException('dirpath应该是文件夹而非文件')
    settings_path = os.path.join(dirpath, filename)
    return settings_path


# 读取配置文件
def read_settings(settings_path: str = None):
    '''
    :param settings_path: 配置文件路径，不填写用默认
    :return: 配置文件字典

        如果没有配置文件，会返回空字典
    '''
    if not settings_path:
        settings_path = get_settings_filepath()
    if not os.path.isfile(settings_path):
        init_settings(settings_path)
    return json.load(open(settings_path, 'rb'))


# 保存配置文件
def save_settings(data: dict, settings_path: str = None):
    '''
    :param data: 配置数据
    :param settings_path: 配置文件路径，不填写用默认
    '''
    if not settings_path:
        settings_path = get_settings_filepath()
    return json.dump(data, open(settings_path, 'w'))


if __name__ == '__main__':
    init_settings()
# if __name__ == '__main__':
#     # print(os.getenv("SystemDrive"))
#     # init_settings()
#     import os
#     # command = f'vi {path}'
#     # os.popen(command)
#     os.system("open -a Terminal .")
#     path = get_settings_filepath()
#     command = f'open {path}'
#     os.system(command)



from typing import Literal, Union
from paux import exception
import math


# 计算手续费
def get_commission_data(
        posSide: Literal['long', 'short'],
        buyMoney: Union[int, float],
        buyLine: Union[int, float],
        sellLine: Union[int, float],
        lever: Union[int, float],
        buyCommissionRate: Union[int, float],
        sellCommissionRate: Union[int, float]
):
    if posSide == 'long':
        buyCommission = buyMoney * lever * buyCommissionRate
        sellCommission = (sellLine / buyLine) * buyMoney * lever * sellCommissionRate
        commission = round(buyCommission + sellCommission, 4)
        sellMoney = round((sellLine - buyLine) / buyLine * lever * buyMoney + buyMoney - commission, 4)
        profitRate = round((sellMoney - buyMoney) / buyMoney, 4)
        return dict(
            sellMoney=sellMoney,
            commission=commission,
            profitRate=profitRate,
        )
    elif posSide == 'short':
        buyCommission = buyMoney * lever * buyCommissionRate
        sellCommission = (2 * buyLine - sellLine) / buyLine * buyMoney * lever * sellCommissionRate
        commission = round(buyCommission + sellCommission, 4)
        sellMoney = (buyLine - sellLine) / buyLine * buyMoney * lever + buyMoney - commission
        profitRate = round((sellMoney - buyMoney) / buyMoney, 4)
        return dict(
            sellMoney=sellMoney,
            commission=commission,
            profitRate=profitRate,
        )


def round_simulate(num):
    if num == 0:
        return num
    px_0_00001 = 0.00001 * num
    if px_0_00001 < 1:
        ndigits = len(str(int(1 / px_0_00001)))
        px_round = round(num, ndigits)
    else:
        px_round = round(num, 1)
    if abs((px_round - num) / num) >= 0.00001:
        raise Exception(
            'px_round={px_round},px={px}'.format(
                px_round=px_round,
                px=num
            )
        )
    else:
        return px_round


def round_floor(num, decimal: int = None, precise: Union[str, int, float] = None):
    if decimal != None:
        if decimal > 0:
            num_floor = math.floor(num * (10 ** decimal)) / (10 ** decimal)
            num_floor = round(num_floor, decimal)
        else:
            num_floor = math.floor(num)
    elif precise != None:
        precise = float(precise)
        decimal = len(str(round(1 / precise) - 1))
        num_floor = math.floor(num / precise) * precise
        if precise >= 1:
            num_floor = round(num_floor)
        else:
            num_floor = round(num_floor, decimal)
        if num_floor > num:
            error_msg = '数字向下取元整出错'
            raise exception.UnexpectError(error_msg)
        if num - num_floor >= precise:
            error_msg = '数字向下取元整出错'
            raise exception.UnexpectError(error_msg)
    else:
        error_msg = 'decimal和precise不能同时为空'
        raise exception.ParamException(error_msg)
    return num_floor


def round_up(num, decimal: int = None, precise: Union[str, int, float] = None):
    if decimal != None:
        if decimal > 0:
            num_up = math.ceil(num * (10 ** decimal)) / (10 ** decimal)
            num_up = round(num_up, decimal)
        else:
            num_up = math.ceil(num)
    elif precise != None:
        precise = float(precise)
        decimal = len(str(round(1 / precise) - 1))
        num_up = math.ceil(num / precise) * precise

        if precise >= 1:
            num_up = round(num_up)
        else:
            num_up = round(num_up, decimal)
        if num_up < num:
            error_msg = '数字向上取元整出错'
            raise exception.UnexpectError(error_msg)
        if num_up - num >= precise:
            error_msg = '数字向上取元整出错'
            raise exception.UnexpectError(error_msg)
    else:
        error_msg = 'decimal和precise不能同时为空'
        raise exception.ParamException(error_msg)
    return num_up


def to_f(num, decimal: int = None, precise: Union[int, float, str] = None, valid: bool = True):
    if decimal != None:
        pass
    elif precise != None:
        precise = float(precise)
        decimal = len(str(round(1 / precise) - 1))
    else:
        error_msg = 'decimal和precise不能同时为空'
        raise exception.ParamException(error_msg)
    fmt = '{:.' + f'{decimal}f' + '}'
    num_str = fmt.format(num)
    if valid:
        if float(num_str) != num:
            error_msg = '验证出错'
            raise exception.UnexpectError(error_msg)
    return num_str




# def to_f(num, valid: bool = True):
#     decimal_max = 30
#     fmt = '{:.' + f'{decimal_max}f' + '}'
#     num_str = fmt.format(num)
#     print(num_str)
#     decimal_str = num_str.split('.')[-1]
#     minus_decimal = 0
#     for s in decimal_str[::-1]:
#         if s == '0':
#             minus_decimal += 1
#         else:
#             break
#     fmt = '{:.' + '{decimal}f'.format(decimal=decimal_max - minus_decimal) + '}'
#     return fmt.format(num)


if __name__ == '__main__':
    import numpy as np

    # a = round_floor(num=123123.67287676767676767, decimal=0)
    # print(a)
    # precise = 11
    # a = round_floor(num=123123.6767676767676767, precise=precise)
    # print(a)
    # print(a / precise)
    # a = round_floor(num=0.1, precise=10)
    # print(a)

    # a = round_up(num=123123.67287676767676767, decimal=0)
    # print(a)
    precise = 0.00000000000000000000007
    a = round_up(num=0.00000000000000000123567, precise=precise)
    print(a)
    print(to_f(a,precise=precise))
    # print(a)
    # a = round_up(num=123123.6767676767676767, precise=0.01)
    # print(a)
    # a = round_up(num=0.1, precise=0.01)



import datetime
import pandas as pd
import numpy as np
from paux import date as _date
from paux.candle import interval as _interval
from typing import Union


# 验证数据间隔
def valid_interval(candle: np.array, interval: Union[int, float] = None, bar: str = None, MINUTE_BAR_INTERVAL=60000):
    if pd.isnull(interval):
        interval = _interval.get_interval(bar=bar, MINUTE_BAR_INTERVAL=MINUTE_BAR_INTERVAL)
    return (np.diff(candle[:, 0]) == interval).all()


# 验证数据的起始时间
def valid_start(candle: np.array, start: Union[int, float, str, datetime.date], timezone: str = None):
    ts = _date.to_ts(date=start, timezone=timezone, default=np.nan)
    if candle[0, 0] == ts:
        return True
    else:
        return False


# 验证数据的终止时间
def valid_end(candle: np.array, end: Union[int, float, str, datetime.date], timezone: str = None):
    ts = _date.to_ts(date=end, timezone=timezone, default=np.nan)
    if candle[-1, 0] == ts:
        return True
    else:
        return False



import numpy as np
from paux import exception

# 估计K线中的时间粒度
def predict_bar(candle: np.array, MINUTE_BAR_INTERVAL=60000) -> str:
    '''
    :param candle: np.array
    :return: bar
    '''
    bar_interval = np.min(np.diff(candle[:, 0]))
    bar_int = bar_interval / MINUTE_BAR_INTERVAL
    if bar_int <= 59:
        suffix = 'm'
        bar_int = bar_int
    elif bar_int <= 60 * 23:
        suffix = 'h'
        bar_int = bar_int / 60
    else:
        bar_int = bar_int / 60 * 24
        suffix = 'd'

    if not bar_int == int(bar_int):
        raise exception.ParamException('Unable to predict bar')
    return '{bar_int}{suffix}'.format(bar_int=int(bar_int), suffix=suffix)


if __name__ == '__main__':
    pass



from threading import Thread
import inspect
import ctypes


def stop_thread(thread):
    try:
        exctype = SystemExit
        tid = ctypes.c_long(thread.ident)
        if not inspect.isclass(exctype):
            exctype = type(exctype)
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
        if res == 1: return True
        else: return False
    except:
        return None

def thread_wrapper(func):
    def wrapper(*args, **kwargs):
        thread_target = Thread(target=func,args=args,kwargs=kwargs)
        thread_target.start()
        return thread_target
    return wrapper


import time


class Filter():
    def __init__(self):
        self.filter_map = {}

    def set(self, name, filter_minute=5):
        ts = int(time.time())
        self.filter_map[name] = {
            'ts': ts,
            'expire': ts + filter_minute * 60
        }

    def check(self, name):
        '''
        :param name:
        :return:  True 不过滤 False 过滤
        '''
        # 没有在过滤内容中
        if not name in self.filter_map.keys():
            return True
        # 未超过过期时间
        if time.time() < self.filter_map[name]['expire']:
            return False
        # 超过了过期时间
        del self.filter_map[name]
        return True

    def clear(self):
        del_keys = []
        for key, v in self.filter_map.items():
            if time.time() >= v['expire']:
                del_keys.append(key)

        for key in del_keys:
            del self.filter_map[key]



import numpy as np
from paux import exception

'''
计算多种金融指标数据
'''


def history_suc(posSide,candle,buyLine,sellLine):
    this_order_index = 0
    suc_num = 0
    if posSide == 'long':
        while True:
            for index in range(this_order_index, candle.shape[0]):
                if candle[index, 3] <= buyLine:
                    this_order_index = index
                    break
                this_order_index = index
            for index in range(this_order_index + 1, candle.shape[0]):
                if candle[index, 2] >= sellLine:
                    this_order_index = index + 1
                    suc_num += 1
                    break
                this_order_index = index
            if this_order_index >= candle.shape[0] - 1:
                break
        return suc_num
    elif posSide == 'short':
        while True:
            for index in range(this_order_index, candle.shape[0]):
                if candle[index, 2] >= buyLine:
                    this_order_index = index
                    break
                this_order_index = index
            for index in range(this_order_index + 1, candle.shape[0]):
                if candle[index, 3] <= sellLine:
                    this_order_index = index + 1
                    suc_num += 1
                    break
                this_order_index = index
            if this_order_index >= candle.shape[0] - 1:
                break
        return suc_num
    else:
        raise exception.ParamException('posSide 应该为long或short')

# 得到均线
def get_candle_ma(
        candle: np.array,
        n: int
):
    '''
    获取历史均线

    :param candle:历史K线数据
    :param n:间隔
    :return:
        array([
                [ts,mac],
                [ts.mac],
                [ts.mac],
            ])
    '''
    # 均线数据
    mac_datas = []
    candle_shape = candle.shape
    for index in range(candle_shape[0]):
        if index < n - 1:
            this_data = [candle[index, 0]] + [np.nan] * (candle_shape[1] - 1)
            mac_datas.append(
                this_data
            )
        else:
            this_data = [candle[index, 0]] + candle[index - n + 1:index + 1, 1:].mean(axis=0).tolist()
            mac_datas.append(
                this_data
            )
    # 均线Candle
    mac_candle = np.array(mac_datas)
    return mac_candle


# 得到布林带
def get_candle_boll(
        candle: np.array,
        n: int
):
    '''
    获取布林带

    :param candle:历史K线数据
    :param n:间隔
    :return:
        array([
            [ts,mb,up,dn],
            [ts,mb,up,dn],
            [ts,mb,up,dn],
        ])
        (mb:均线 up:上轨 dn:下轨)
    '''
    # 布林带数据
    boll_datas = []
    for index in range(candle.shape[0]):
        if index <= n - 1:
            mb = np.nan
            up = np.nan
            dn = np.nan
        else:
            mb = candle[index - n:index, 4].mean()  # 收盘价均值
            sd = (candle[index - n:index, 4] - mb).std()  # 收盘价标准差
            up = mb + 2 * sd  # 上轨
            dn = mb - 2 * sd  # 下轨
        boll_datas.append(
            [candle[index, 0], mb, up, dn]
        )
    # 布林带Candle
    boll_candle = np.array(boll_datas)
    return boll_candle


# 得到DualThrust中的上轨和下轨
def get_candle_dualThrust(
        candle: np.array,
        n: int,
        ks: float,
        kx: float,
        update_n: int = None
):
    '''
    获取DualThrust Candle

    :param candle:历史K线数据
    :param n:间隔
    :param ks:ks参数
    :param kx:kx参数
    :param update_n:更新的间隔n
    :return:
        array([
            [ts,range_ks,range_kx],
            [ts,range_ks,range_kx],
            [ts,range_ks,range_kx],
        ])
        (range_ks上轨 range_kx下轨)
    '''
    # dualThrust数据
    dualThrust_datas = []
    for i in range(0, candle.shape[0], update_n):
        if i <= n - 1:
            range_kx = np.nan
            range_ks = np.nan
        else:
            last_k_array = candle[i - n:i]  # 前n间隔的数据
            hh = last_k_array[:, 2].max()  # 最高价的最高价
            hc = last_k_array[:, 4].max()  # 最高价的收盘价
            lc = last_k_array[:, 4].min()  # 最低价的收盘价
            ll = last_k_array[:, 3].min()  # 最低价的最低价
            range_ = max(hh - lc, hc - ll)
            o = candle[i, 1]
            range_ks = o + ks * range_  # 上轨
            range_kx = o - kx * range_  # 下轨
        for ts in candle[i:i + update_n, 0]:
            dualThrust_datas.append(
                [ts, range_ks, range_kx]
            )
    # dualThrust Candle
    dualThrust_candle = np.array(dualThrust_datas)
    return dualThrust_candle



'''
TODO 加入log日志
解决multiprocessing中Pool的不稳定性
'''
from multiprocessing import Process, Manager
import traceback
import time
from collections.abc import Iterable
from copy import deepcopy

def _pool_worker(q_param, q_result):
    while True:
        try:
            data = q_param.get(block=False, timeout=0)
            index = data['index']
            param = data['param']
            skip_exception = data['skip_exception']
            func = param['func']
            del param['func']
        except:
            break
        if skip_exception:
            try:
                ret = func(**param)
                q_result.put(
                    {
                        'index': index,
                        'data': ret
                    }
                )
            except:
                print(traceback.format_exc())
        else:
            ret = func(**param)
            q_result.put(
                {
                    'index': index,
                    'data': ret
                }
            )


def pool_worker(params, p_num=4, func=None, skip_exception=False):
    results = [None] * len(params)
    # 标准化param，让param中存在要执行的函数func
    q_param = Manager().Queue()
    q_result = Manager().Queue()
    for index, param in enumerate(params):
        # param['func']优先级最高
        # 存在    param['func']
        if 'func' in param.keys() and param['func'] != None:
            pass
        # 不存在   param['func']
        else:
            # 有func
            if func != None:
                param['func'] = func
            # 无func
            else:
                raise Exception('没有执行的函数func')  # todo
        q_param.put(
            {
                'index': index,
                'param': param,
                'skip_exception': skip_exception
            }
        )
    # 单进程运行
    if p_num <= 1:
        _pool_worker(q_param, q_result)
    # 多进程运行
    else:
        processes = []
        for i in range(p_num):
            p = Process(target=_pool_worker, kwargs={'q_param': q_param, 'q_result': q_result})
            processes.append(p)
            p.start()

        def wait_processes(processes):
            for p in processes:
                if p.is_alive():
                    return False
            return True

        # 等待进程均运行完成
        while True:
            if wait_processes(processes):
                break
            else:
                time.sleep(1)
    # 整理结果
    for i in range(q_result.qsize()):
        result = q_result.get(block=False, timeout=0)
        index = result['index']
        data = result['data']
        # 按照索引赋值，如果某个参数执行异常并且skip_exception，这个结果为None
        results[index] = data
    return results







import logging
import os
import datetime
import traceback


class Log():
    LOG_DIRPATH = './LOG_DATA'
    FILE_LEVEL = 'DEBUG'
    CONSOLE_LEVEL = 'DEBUG'

    def __init__(self):
        if not os.path.isdir(self.LOG_DIRPATH):
            os.makedirs(self.LOG_DIRPATH)

    def get_logger(self, level):
        log_filename = datetime.datetime.now().strftime('%Y-%m-%d') + '.log'
        log_filepath = os.path.join(self.LOG_DIRPATH, log_filename)
        file_handler = logging.FileHandler(log_filepath)  # 输出到文件
        console_handler = logging.StreamHandler()  # 输出到控制台
        file_handler.setLevel(self.FILE_LEVEL)  # FILE_LEVEL以上才输出到文件
        console_handler.setLevel(self.CONSOLE_LEVEL)  # CONSOLE_LEVEL以上才输出到控制台
        fmt = '''-----------------------------------------------
        datetime:   %(asctime)s
        level:      %(levelname)s
        filePath:   {filepath}
        funcName:   %(funcName)s
        line:       %(lineno)d
        messge:     {message}
        traceback:  {traceback}
        '''.format(filepath=__file__, message='%(message)s', traceback=str(traceback.format_exc()))
        formatter = logging.Formatter(fmt)
        file_handler.setFormatter(formatter)  # 设置文件内容格式
        console_handler.setFormatter(formatter)  # 设置控制台内容格式
        logger = logging.getLogger('updateSecurity')
        logger.setLevel(level)
        logger.addHandler(file_handler)
        logger.addHandler(console_handler)
        return logger

    def log_info(self, msg):
        LEVEL = 'INFO'
        self.get_logger(level=LEVEL).info(msg)

    def log_debug(self, msg):
        LEVEL = 'DEBUG'
        self.get_logger(level=LEVEL).info(msg)

    def log_error(self, msg):
        LEVEL = 'ERROR'
        self.get_logger(level=LEVEL).info(msg)

    def log_warn(self, msg):
        LEVEL = 'warn'
        self.get_logger(level=LEVEL).info(msg)

# if __name__ == '__main__':
#     l = Log()
#     l.log_info(msg='123')
